import pytest

import acurl_ng


# Curl will return a non-zero exit code when something goes wrong
# https://curl.se/libcurl/c/libcurl-errors.html
# https://everything.curl.dev/usingcurl/returns
@pytest.mark.asyncio
async def test_non_zero_exit_code_raises(acurl_session_ng):
    with pytest.raises(
        acurl_ng.AcurlError, match="curl failed with code 6 Couldn't resolve host name"
    ):
        # Curl returns CURLE_COULDNT_RESOLVE_HOST (6) exit code
        await acurl_session_ng.get("unresolvable_hostname.doesnotexist")



import asyncio
import gc

import acurl_ng


async def main():
    el = acurl_ng.CurlWrapper(asyncio.get_running_loop())
    session = el.session()

    # Sometimes a leak won't be apparent when running a line of code just
    # once, due to the vagaries of gc etc.  A useful tactic is to run a
    # possibly leaky operation 100 times, and then look for things in the
    # valgrind output that are lost or possibly lost 100 (or possibly 99)
    # times -- that's an indication the allocations come from the code run in
    # the loop.

    # for _ in range(100):
    #     r = await session.get("http://example.com")

    # But the simplest case is just to run the code once
    await session.get("http://example.com")


if __name__ == "__main__":
    asyncio.run(main())
    # Run the GC before exiting just to ensure that all possible collections
    # are done.
    gc.collect()



import acurl_ng


def create_request(
    method,
    url,
    headers=(),
    cookies=(),
    auth=None,
    data=None,
    cert=None,
):
    # Cookies should be the byte string representation of the cookie
    if isinstance(method, str):
        method = method.encode()
    headers = tuple(h.encode("utf-8") if hasattr(h, "encode") else h for h in headers)
    return acurl_ng.Request(method, url, headers, cookies, auth, data, cert)



import pytest
from helpers_ng import create_request

import acurl_ng


def test_to_curl():
    r = create_request("GET", "http://foo.com")
    assert r.to_curl() == "curl -X GET http://foo.com"


def test_to_curl_headers():
    r = create_request(
        "GET", "http://foo.com", headers=("Foo: bar", "My-Header: is-awesome")
    )
    assert (
        r.to_curl()
        == "curl -X GET -H 'Foo: bar' -H 'My-Header: is-awesome' http://foo.com"
    )


def test_to_curl_cookies():
    r = create_request(
        "GET",
        "http://foo.com",
        cookies=(
            acurl_ng._Cookie(
                False, "foo.com", True, "/", False, 0, "123", "456"
            ).format(),
        ),
    )
    assert r.to_curl() == "curl -X GET --cookie 123=456 http://foo.com"


def test_to_curl_multiple_cookies():
    r = create_request(
        "GET",
        "http://foo.com",
        cookies=(
            acurl_ng._Cookie(
                False, "foo.com", True, "/", False, 0, "123", "456"
            ).format(),
            acurl_ng._Cookie(
                False, "foo.com", True, "/", False, 0, "789", "abc"
            ).format(),
        ),
    )
    assert r.to_curl() == "curl -X GET --cookie '123=456;789=abc' http://foo.com"


# FIXME: curl won't send cookies that don't match the domain.
# https://curl.se/docs/http-cookies.html
# Should we ignore cookies that won't be sent?
@pytest.mark.skip(reason="unimplemented")
def test_to_curl_cookies_wrong_domain():
    # I'm not sure if this is a valid test case...Request objects should
    # probably only be constructed via Session.request, which always creates
    # cookies for the domain of the request.  So the case this is exercising
    # won't ever happen.
    r = create_request(
        "GET",
        "http://foo.com",
        # The domain doesn't match, the cookie should not be passed
        cookies=(
            acurl_ng._Cookie(
                False, "bar.com", True, "/", False, 0, "123", "456"
            ).format(),
        ),
    )
    assert r.to_curl() == "curl -X GET http://foo.com"


def test_to_curl_auth():
    r = create_request("GET", "http://foo.com", auth=("user", "pass"))
    assert r.to_curl() == "curl -X GET --user user:pass http://foo.com"



#cython: language_level=3

from libc.stdlib cimport malloc
from libc.string cimport strndup
from cpython.pycapsule cimport PyCapsule_New, PyCapsule_GetPointer
from curlinterface cimport *
from cpython.ref cimport Py_INCREF
from libc.stdio cimport printf
from libc.time cimport time
import cython
from json import dumps

class RequestError(Exception):
    pass

# Callback functions

cdef BufferNode* alloc_buffer_node(size_t size, char *data):
    cdef BufferNode* node = <BufferNode*>malloc(sizeof(BufferNode))
    if node == NULL:
        printf("OOOPS!!!!")  # FIXME: better checks
    node.len = size
    # FIXME: the curl docs give an example function that uses realloc on a
    # single buffer rather than a linked list.  Possibly that method would be
    # faster (as well as less complicated) -- especially if we preallocate a
    # buffer for headers/data when allocating a response (paying a small
    # memory cost in exchange for not having to allocate while processing
    # response data).  It would also make the code less complicated.  Worth
    # benchmarking the effects someday...
    # <https://curl.se/libcurl/c/CURLOPT_WRITEFUNCTION.html>
    node.buffer = strndup(data, size)
    node.next = NULL
    return node

cdef size_t header_callback(char *ptr, size_t size, size_t nmemb, void *userdata):
    cdef _Response response = <_Response>userdata
    cdef BufferNode* node = alloc_buffer_node(size * nmemb, ptr)
    if response.header_buffer == NULL:  # FIXME: unlikely
        response.header_buffer = node
    if response.header_buffer_tail != NULL:  # FIXME: likely
        response.header_buffer_tail.next = node
    response.header_buffer_tail = node
    return node.len

cdef size_t body_callback(char *ptr, size_t size, size_t nmemb, void *userdata):
    cdef _Response response = <_Response>userdata
    cdef BufferNode* node = alloc_buffer_node(size * nmemb, ptr)
    if response.body_buffer == NULL:  # FIXME: unlikely
        response.body_buffer = node
    if response.body_buffer_tail != NULL:  # FIXME: likely
        response.body_buffer_tail.next = node
    response.body_buffer_tail = node
    return node.len

cdef void cleanup_share(object share_capsule):
    cdef void* share_raw = PyCapsule_GetPointer(share_capsule, <const char*>NULL)
    curl_share_cleanup(<CURLSH*>share_raw)

# From the cython docs
# <https://cython.readthedocs.io/en/latest/src/userguide/extension_types.html#disabling-cycle-breaking-tp-clear>:
# > If any Python objects can be referenced [by this class], Cython will
# > automatically generate the tp_traverse and tp_clear slots.  There is at
# > least one reason why this might not be what you want: if you need to cleanup
# > some external resources in the __dealloc__ special function and your object
# > happened to be in a reference cycle, the garbage collector may have
# > triggered a call to tp_clear to clear the object.  In that case, any object
# > references have vanished when __dealloc__ is called.
#
# This is exactly the situation here, and in previous interations of this code
# we got some nasty crashes (segfaults) without this decorator.  It's
# theoretically dangerous (could leak memory), but as the cython docs say:
#
# > If you use no_gc_clear, it is important that any given reference cycle
# > contains at least one object without no_gc_clear.
#
# Since this is the only class that has this decorator, we're fine (a
# reference cycle will never consist solely of Session objects referencing
# other Sessions) -- tp_clear will be called on some other object in the cycle.
@cython.no_gc_clear
cdef class Session:
    cdef CURLSH* shared
    cdef CurlWrapper wrapper
    cdef public object response_callback

    def __cinit__(self, wrapper):
        self.shared = curl_share_init()
        acurl_share_setopt_int(self.shared, CURLSHOPT_SHARE, CURL_LOCK_DATA_COOKIE)
        acurl_share_setopt_int(self.shared, CURLSHOPT_SHARE, CURL_LOCK_DATA_DNS)
        acurl_share_setopt_int(self.shared, CURLSHOPT_SHARE, CURL_LOCK_DATA_SSL_SESSION)
        self.wrapper = wrapper
        self.response_callback = None

    def __dealloc__(self):
        if self.wrapper.loop is None or self.wrapper.loop.is_closed():
            # FIXME: the event loop being closed only happens during testing,
            # so it kind of sucks to pay the price of checking for it all the
            # time.  I'm not even sure what the circumstances are under which
            # self.wrapper.loop becomes None -- I suspect it has to do with
            # the cycle collector.
            curl_share_cleanup(self.shared)
        else:
            self.wrapper.loop.call_soon(cleanup_share, PyCapsule_New(self.shared, NULL, NULL))

    def cookies(self):
        cdef CURL* curl = curl_easy_init()
        acurl_easy_setopt_voidptr(curl, CURLOPT_SHARE, self.shared)
        lst = acurl_extract_cookielist(curl)
        curl_easy_cleanup(curl)
        return cookie_seq_to_cookie_dict(tuple(parse_cookie_string(c) for c in lst))

    def erase_all_cookies(self):
        cdef CURL* curl = curl_easy_init()
        acurl_easy_setopt_voidptr(curl, CURLOPT_SHARE, self.shared)
        acurl_easy_setopt_cstr(curl, CURLOPT_COOKIELIST, b"ALL")
        curl_easy_cleanup(curl)

    cdef object _inner_request(
        self,
        bytes method,
        str url,
        tuple headers,
        tuple cookies,
        object auth,
        bytes data,
        object cert,
    ):
        cdef curl_slist* curl_headers = NULL
        cdef CURL* curl = curl_easy_init()
        cdef object future = self.wrapper.loop.create_future()

        acurl_easy_setopt_voidptr(curl, CURLOPT_SHARE, self.shared)
        acurl_easy_setopt_cstr(curl, CURLOPT_URL, url.encode())
        acurl_easy_setopt_cstr(curl, CURLOPT_CUSTOMREQUEST, method)
        # curl_easy_setopt(rd->curl, CURLOPT_VERBOSE, 1)  # DEBUG
        acurl_easy_setopt_cstr(curl, CURLOPT_ENCODING, b"")
        # FIXME: make this configurable?
        acurl_easy_setopt_int(curl, CURLOPT_SSL_VERIFYPEER, 0)
        acurl_easy_setopt_int(curl, CURLOPT_SSL_VERIFYHOST, 0)

        cdef int i

        for i in range(len(headers)):
            if not isinstance(headers[i], bytes):
                raise ValueError("headers should be a tuple of bytestrings if set")
            curl_headers = curl_slist_append(curl_headers, headers[i])
        acurl_easy_setopt_voidptr(curl, CURLOPT_HTTPHEADER, curl_headers)

        # We have to postpone the initialization of the Request object until
        # here, because we're going to transfer the ownership of the
        # curl_headers to it
        cdef Request request = Request.__new__(Request, method, url, headers, cookies, auth, data, cert)
        request.store_session_cookies(self.shared)
        request.curl_headers = curl_headers
        cdef _Response response = _Response.make(self, curl, future, time(NULL), request)

        acurl_easy_setopt_voidptr(curl, CURLOPT_PRIVATE, <void*>response)
        acurl_easy_setopt_writecb(curl, CURLOPT_WRITEFUNCTION, body_callback)
        acurl_easy_setopt_voidptr(curl, CURLOPT_WRITEDATA, <void*>response)
        acurl_easy_setopt_writecb(curl, CURLOPT_HEADERFUNCTION, header_callback)
        acurl_easy_setopt_voidptr(curl, CURLOPT_HEADERDATA, <void*>response)

        # We need to increment the reference count on the response.  To
        # python's eyes the last reference to it disappears when we exit this
        # function (and thus it would get collected) -- but the curl event
        # loop has a reference to it so it needs to remain alive.  FIXME:
        # where does it get decrefed?  In principle we need a matching
        # Py_DECREF somewhere in the code.  But there's malignant
        # counter-wizardry coming from cython -- when we coerce the pointer
        # out of curl, it gets assigned into a python variable, which cython
        # will "helpfully" decref for us when it goes out of scope -- so the
        # decref might already be handled for us (we do prevent the magic
        # decref in some circumstances, though).  This is not really an
        # optimal situation and we need to get to the bottom of it...  (Is it
        # related to the no_gc_clear and the crashes that it now prevents?)
        Py_INCREF(response)

        if auth is not None:
            if (
                not isinstance(auth, tuple)
                or len(auth) != 2
                or not isinstance(auth[0], str)
                or not isinstance(auth[1], str)
            ):
                raise ValueError(f"auth must be a 2-tuple of strings but it was: {auth}")
            acurl_easy_setopt_cstr(curl, CURLOPT_USERPWD, auth[0].encode() + b":" + auth[1].encode())

        if cert is not None:
            if (
                not isinstance(cert, tuple)
                or len(cert) != 2
                or not isinstance(cert[0], str)
                or not isinstance(cert[1], str)
            ):
                raise ValueError(f"cert must be a 2-tuple of strings but it was: {cert}")
            acurl_easy_setopt_cstr(curl, CURLOPT_SSLKEY, cert[0])
            acurl_easy_setopt_cstr(curl, CURLOPT_SSLCERT, cert[1])

        acurl_easy_setopt_cstr(curl, CURLOPT_COOKIEFILE, "")
        if cookies is not None:
            for i in range(len(cookies)):
                if not isinstance(cookies[i], bytes):
                    raise ValueError("cookies should be a tuple of bytes")
                acurl_easy_setopt_cstr(curl, CURLOPT_COOKIELIST, cookies[i])

        if data is not None:
            acurl_easy_setopt_int(curl, CURLOPT_POSTFIELDSIZE, len(data))
            acurl_easy_setopt_cstr(curl, CURLOPT_POSTFIELDS, data)

        curl_multi_add_handle(self.wrapper.multi, curl)

        return future

    async def _outer_request(
        self,
        method,
        url,
        *,
        headers=(),
        cookies=None,
        auth=None,
        data=None,
        json=None,
        cert=None,
        allow_redirects=True,
        max_redirects=5,
    ):
        if not isinstance(method, bytes):
            raise ValueError("method must be bytes")
        if not isinstance(url, str):
            raise ValueError("url must be a string")
        headers = dict(headers)
        if cookies is None:
            cookies = {}
        if json is not None:
            if data is not None:
                raise ValueError("use only one or none of data or json")
            data = dumps(json).encode()  # FIXME: make fast
            headers.setdefault("Content-Type", "application/json")
        elif data is not None:
            if isinstance(data, str):
                data = data.encode()
            else:
                if not isinstance(data, bytes):
                    # FIXME: not very duck type
                    raise ValueError("data must be str or bytes")

        # FIXME: probably need to do some escaping if one of the values has
        # newlines in it...
        cdef tuple headers_tuple = tuple(k.encode() + b": " + v.encode() for k, v in headers.items())
        cdef tuple cookie_tuple = tuple(session_cookie_for_url(url, k, v).format() for k, v in cookies.items())

        cdef _Response response = await self._inner_request(
            method,
            url,
            headers_tuple,
            cookie_tuple,
            auth,
            data,
            cert,
        )
        # This decref is the partner to the incref in _inner_request above --
        # at this point curl is done with the response and so we no longer
        # need to keep its refcount incremented to prevent the GC from
        # cleaning it up when the only reference to it is held by curl and not
        # python.
        Py_DECREF(response)
        cdef _Response old_response
        if self.response_callback:
            # FIXME: should this be async?
            self.response_callback(response)
        if (allow_redirects and (300 <= response.status_code < 400) and response.redirect_url is not None):
            while (max_redirects > 0):
                max_redirects -= 1
                if response.status_code in {301, 302, 303}:
                    method = b"GET"
                    data = None
                old_response = response
                response = await self._inner_request(
                    method,
                    response.redirect_url,
                    headers_tuple,
                    cookie_tuple,
                    auth,
                    data,
                    cert,
                )
                if self.response_callback:
                    # FIXME: should this be async?
                    self.response_callback(response)
                response._set_prev(old_response)
                if not ((300 <= response.status_code < 400)
                        and response.redirect_url is not None):
                    break
            else:
                raise RequestError("Max Redirects")
        return response

    def set_response_callback(self, callback):
        self.response_callback = callback

    async def delete(self, *args, **kwargs):
        return await self._outer_request(b"DELETE", *args, **kwargs)

    async def get(self, *args, **kwargs):
        return await self._outer_request(b"GET", *args, **kwargs)

    async def head(self, *args, **kwargs):
        return await self._outer_request(b"HEAD", *args, **kwargs)

    async def options(self, *args, **kwargs):
        return await self._outer_request(b"OPTIONS", *args, **kwargs)

    async def patch(self, *args, **kwargs):
        return await self._outer_request(b"PATCH", *args, **kwargs)

    async def post(self, *args, **kwargs):
        return await self._outer_request(b"POST", *args, **kwargs)

    async def put(self, *args, **kwargs):
        return await self._outer_request(b"PUT", *args, **kwargs)



#cython: language_level=3

from libc.stdlib cimport free
from cpython cimport array
from json import loads
from cpython.list cimport PyList_New
from libc.stdio cimport printf
import warnings

cdef struct BufferNode:
    size_t len
    char *buffer
    BufferNode *next

cdef list acurl_extract_cookielist(CURL* curl):
    cdef void* start_raw
    cdef curl_slist *start
    cdef curl_slist *node
    acurl_easy_getinfo_voidptr(curl, CURLINFO_COOKIELIST, &start_raw)
    start = <curl_slist*>start_raw
    node = start
    cdef lst = []
    while node != NULL:
        lst.append(node.data.decode("UTF-8"))
        node = node.next
    curl_slist_free_all(start)
    return lst

cdef class _Response:
    cdef BufferNode* header_buffer
    cdef BufferNode* header_buffer_tail
    cdef BufferNode* body_buffer
    cdef BufferNode* body_buffer_tail
    cdef CURL* curl
    cdef Session session
    cdef object future
    cdef readonly unsigned long start_time
    cdef readonly Request request
    cdef _Response _prev

    def __cinit__(self):
        # Technically it's dangerous to leave curl and session members
        # uninitialized, but we hope no one calls our init method directly...
        self.header_buffer = NULL
        self.header_buffer_tail = NULL
        self.body_buffer = NULL
        self.body_buffer_tail = NULL
        self._prev = None

    def __dealloc__(self):
        # Subtle behavior alert!  We cleanup the curl handle before we free
        # header_buffer and body_buffer.  As the curl docs say:
        # > Occasionally you may get your progress callback or header callback
        # > called from within curl_easy_cleanup.
        # If that happens, then the handler would attach more nodes to the
        # header/body buffers, which we'll need to free.  That is perhaps
        # unlikely, given that the docs say further:
        # > [This happens if] the protocol is of a kind that requires a
        # > command/response sequence before disconnect.  Examples of such
        # > protocols are FTP, POP3 and IMAP.
        # But better safe than sorry.
        curl_easy_cleanup(self.curl)
        cdef BufferNode* ptr
        cdef BufferNode* old_ptr
        ptr = self.header_buffer
        while ptr != NULL:
            old_ptr = ptr
            free(ptr.buffer)
            ptr = ptr.next
            free(old_ptr)
        ptr = self.body_buffer
        while ptr != NULL:
            old_ptr = ptr
            free(ptr.buffer)
            ptr = ptr.next
            free(old_ptr)

    # In principle it would be better to do this through a normal init
    # method.  Howver, there is a restriction on what arguments can be passed
    # to an init method -- they must all be python types (C pointers don't
    # work).  So we by convention say that you shouldn't do `r =
    # _Response(...)` but rather only ever write `r = _Response.make(...)`.
    # (There's no concrete enforcement of this requirement, other than "tests
    # will probably break if you disobey".)  An alternative might be to move
    # lots of the initialization of the CURL* from _Session._inner_request
    # into this class, and lean into the idea that it's _Response that owns
    # the CURL*.  Then _inner_request would shrink quite a bit, only being
    # responsible for creating the request, creating the response, and then
    # submitting the response's CURL* to the multi.
    @staticmethod
    cdef _Response make(Session session, CURL* curl, object future, unsigned long time, Request request):
        cdef _Response r = _Response.__new__(_Response)
        r.session = session
        r.curl = curl
        r.future = future
        r.start_time = time
        r.request = request
        return r

    cdef long get_info_long(self, CURLINFO info):
        cdef long value
        acurl_easy_getinfo_long(self.curl, info, &value)
        return value

    cdef str get_info_str(self, CURLINFO info):
        cdef char* value
        acurl_easy_getinfo_cstr(self.curl, info, &value)
        return value.decode('UTF-8')

    cdef double get_info_double(self, CURLINFO info):
        cdef double value
        acurl_easy_getinfo_double(self.curl, info, &value)
        return value

    cdef list get_cookielist(self):
        return acurl_extract_cookielist(self.curl)

    def _set_prev(self, prev):
        self._prev = prev

    @property
    def status_code(self):
        return self.get_info_long(CURLINFO_RESPONSE_CODE)

    @property
    def response_code(self):
        warnings.warn(
            "Deprecated: Please consider using the status_code method instead",
            DeprecationWarning,
            stacklevel=2
        )
        return self.get_info_long(CURLINFO_RESPONSE_CODE)

    @property
    def url(self):
        return self.get_info_str(CURLINFO_EFFECTIVE_URL)

    @property
    def redirect_url(self):
        return self.get_info_str(CURLINFO_REDIRECT_URL)

    @property
    def total_time(self):
        return self.get_info_double(CURLINFO_TOTAL_TIME)

    @property
    def namelookup_time(self):
        return self.get_info_double(CURLINFO_NAMELOOKUP_TIME)

    @property
    def connect_time(self):
        return self.get_info_double(CURLINFO_CONNECT_TIME)

    @property
    def appconnect_time(self):
        return self.get_info_double(CURLINFO_APPCONNECT_TIME)

    @property
    def pretransfer_time(self):
        return self.get_info_double(CURLINFO_PRETRANSFER_TIME)

    @property
    def starttransfer_time(self):
        return self.get_info_double(CURLINFO_STARTTRANSFER_TIME)

    @property
    def upload_size(self):
        return self.get_info_double(CURLINFO_SIZE_UPLOAD)

    @property
    def download_size(self):
        return self.get_info_double(CURLINFO_SIZE_DOWNLOAD)

    @property
    def primary_ip(self):
        return self.get_info_str(CURLINFO_PRIMARY_IP)

    @property
    def cookies(self):
        # FIXME: we lose the ability to get domains of cookies...
        return cookie_seq_to_cookie_dict(tuple(parse_cookie_string(c) for c in self.get_cookielist()))

    @property
    def history(self):
        cdef list result = []
        cdef _Response cur = self._prev
        while cur is not None:
            result.append(cur)
            cur = cur._prev
        result.reverse()
        return result

    @property
    def body(self):
        cdef array.array body = array.array("B")
        cdef BufferNode* node = self.body_buffer
        while node != NULL:
            array.extend_buffer(body, node.buffer, node.len)
            node = node.next
        return body.tobytes()

    # TODO: is this part of the request api?
    @property
    def header(self):
        # Idea to cache this: need three states:
        # - headers not populated: header_buffer is null, header_buffer_tail
        # is not
        # - headers being populated: both not null
        # - headers calculated: header_buffer is not null (and points to
        # array); tail is null
        # then make all the code behave correctly
        cdef array.array header = array.array("B")
        cdef BufferNode* node = self.header_buffer
        while node != NULL:
            array.extend_buffer(header, node.buffer, node.len)
            node = node.next
        return header.tobytes().decode("UTF-8")

    @property
    def encoding(self):
        if "Content-Type" in self.headers and "charset=" in self.headers["Content-Type"]:
            return self.headers["Content-Type"].split("charset=")[-1].split()[0]
        return "latin1"

    @property
    def text(self):
        return self.body.decode(self.encoding)

    def json(self):
        return loads(self.body)

    @property
    def headers(self):
        cdef object headers_pre = CaseInsensitiveDefaultDict(list)
        cdef str k, v
        for k, v in self.headers_tuple:
            headers_pre[k].append(v)
        cdef object headers = CaseInsensitiveDict()
        for k in headers_pre:
            headers[k] = ", ".join(headers_pre[k])
        return headers  # FIXME: should be frozen

    def _get_header_lines(self):
        cdef list headers = self.header.split("\r\n")
        headers = headers[:-2]  # drop the final blank lines
        while headers[0].startswith("HTTP/1.1 100"):
            headers = headers[2:]
        return headers[1:]  # drop the final response code

    # TODO: is this part of the request api?
    @property
    def headers_tuple(self):
        return tuple(tuple(line.split(": ", 1)) for line in self._get_header_lines())



import pytest
from werkzeug.wrappers import Response as WZResponse

import acurl_ng


@pytest.mark.asyncio
async def test_request_no_data_raises(acurl_session_ng):
    with pytest.raises(ValueError):
        await acurl_session_ng.get("http://foo.com", json={}, data="")


@pytest.mark.asyncio
async def test_get(httpserver, acurl_session_ng):
    httpserver.expect_oneshot_request("/test", "GET").respond_with_data("hi")
    r = await acurl_session_ng.get(httpserver.url_for("/test"))

    assert isinstance(r, acurl_ng._Response)
    assert r.request.method == b"GET"
    assert r.body == b"hi"
    httpserver.check_assertions()


@pytest.mark.asyncio
async def test_get_headers(httpserver, acurl_session_ng):
    httpserver.expect_oneshot_request(
        "/test", "GET", headers={"My-Header": "is-awesome"}
    ).respond_with_data("hi")
    r = await acurl_session_ng.get(
        httpserver.url_for("/test"), headers={"My-Header": "is-awesome"}
    )

    assert isinstance(r, acurl_ng._Response)
    assert r.request.method == b"GET"
    assert r.body == b"hi"
    httpserver.check_assertions()


@pytest.mark.asyncio
async def test_post_json(httpserver, acurl_session_ng):
    httpserver.expect_oneshot_request(
        "/test",
        "POST",
        json={"foo": "bar"},
        headers={"content-type": "application/json"},
    ).respond_with_data("hi")

    r = await acurl_session_ng.post(httpserver.url_for("/test"), json={"foo": "bar"})

    assert isinstance(r, acurl_ng._Response)
    assert r.request.method == b"POST"
    assert r.body == b"hi"
    httpserver.check_assertions()


@pytest.mark.asyncio
async def test_post_data(httpserver, acurl_session_ng):
    httpserver.expect_oneshot_request("/test", "POST", data="foobar").respond_with_data(
        "hi"
    )

    r = await acurl_session_ng.post(httpserver.url_for("/test"), data="foobar")

    assert isinstance(r, acurl_ng._Response)
    assert r.request.method == b"POST"  # FIXME: should it be a string?
    assert r.body == b"hi"
    httpserver.check_assertions()


@pytest.mark.asyncio
async def test_response_callback(httpserver, acurl_session_ng):
    httpserver.expect_oneshot_request("/test", "GET").respond_with_data("hi")
    called = False

    def response_cb(resp):
        nonlocal called
        if called:
            raise Exception("called too many times")
        assert isinstance(resp, acurl_ng._Response)
        called = True

    acurl_session_ng.response_callback = response_cb
    await acurl_session_ng.get(httpserver.url_for("/test"))
    assert called
    httpserver.check_assertions()


@pytest.mark.asyncio
async def test_redirect(httpserver, acurl_session_ng):
    httpserver.expect_oneshot_request("/test", "GET").respond_with_response(
        WZResponse(status=301, headers={"Location": "/test2"})
    )
    httpserver.expect_oneshot_request("/test2", "GET").respond_with_data("hi")

    r = await acurl_session_ng.get(httpserver.url_for("/test"))

    assert r.body == b"hi"
    assert r.url == httpserver.url_for("/test2")
    h = r.history
    assert len(h) == 1
    assert h[0].url == httpserver.url_for("/test")


@pytest.mark.asyncio
async def test_max_redirects_raises(httpserver, acurl_session_ng):
    httpserver.expect_oneshot_request("/test", "GET").respond_with_response(
        WZResponse(status=301, headers={"Location": "/test2"})
    )

    with pytest.raises(acurl_ng.RequestError):
        await acurl_session_ng.get(httpserver.url_for("/test"), max_redirects=0)


@pytest.mark.asyncio
async def test_disallow_redirects(httpserver, acurl_session_ng):
    httpserver.expect_oneshot_request("/test", "GET").respond_with_response(
        WZResponse(status=301, headers={"Location": "/test2"})
    )

    r = await acurl_session_ng.get(httpserver.url_for("/test"), allow_redirects=False)

    assert r.status_code == 301
    assert r.headers["Location"].endswith("/test2")



#cython: language_level=3

from curlinterface cimport *
from libc.stdio cimport printf
from cpython cimport array
from cpython.ref cimport Py_DECREF

include "utils.pyx"
include "cookie.pyx"
include "request.pyx"
include "response.pyx"
include "session.pyx"

class AcurlError(Exception):
    pass

# Callback functions

# In the classic (old) acurl interface, the socket and timer functions didn't
# take the GIL (and didn't need to).  However, now because we are cooperating
# with the python asyncio library we need to call back into python.  So these
# functions do take the GIL, as indicated by `with gil` in their signatures.
# The performance impact of this is yet tbd.
cdef int handle_socket(CURL *easy, curl_socket_t sock, int action, void *userp, void *socketp) with gil:
    cdef CurlWrapper wrapper = <CurlWrapper>userp
    if action == CURL_POLL_IN or action == CURL_POLL_INOUT:
        wrapper.loop.add_reader(sock, wrapper.curl_perform_read, wrapper, sock)
    if action == CURL_POLL_OUT or action == CURL_POLL_INOUT:
        wrapper.loop.add_writer(sock, wrapper.curl_perform_write, wrapper, sock)
    if action == CURL_POLL_REMOVE:
        wrapper.loop.remove_reader(sock)
        wrapper.loop.remove_writer(sock)
    if action != CURL_POLL_IN and action != CURL_POLL_OUT and action != CURL_POLL_INOUT and action != CURL_POLL_REMOVE:
        raise Exception("oops")

cdef int start_timeout(CURLM *multi, long timeout_ms, void *userp) with gil:
    cdef CurlWrapper wrapper = <CurlWrapper>userp
    cdef int _running
    cdef double secs
    if timeout_ms < 0:
        if wrapper.timer_handle is not None:
            wrapper.timer_handle.cancel()
            wrapper.timer_handle = None
    elif timeout_ms == 0:
        with nogil:
            curl_multi_socket_action(multi, CURL_SOCKET_TIMEOUT, 0, &_running)
        wrapper.loop.call_soon(wrapper.check_multi_info, wrapper)  # FIXME: are we sure we're on the main thread?
    else:
        secs = timeout_ms / 1000
        wrapper.timer_handle = wrapper.loop.call_later(secs, wrapper.timeout_expired, wrapper)

cdef class CurlWrapper:
    cdef CURLM* multi
    cdef object timer_handle
    cdef object loop

    def __cinit__(self, object loop):
        self.multi = curl_multi_init()
        acurl_multi_setopt_long(self.multi, CURLMOPT_MAXCONNECTS, 1000)  # FIXME: magic number
        acurl_multi_setopt_socketcb(self.multi, CURLMOPT_SOCKETFUNCTION, handle_socket)
        acurl_multi_setopt_pointer(self.multi, CURLMOPT_SOCKETDATA, <void*>self)
        acurl_multi_setopt_timercb(self.multi, CURLMOPT_TIMERFUNCTION, start_timeout)
        acurl_multi_setopt_pointer(self.multi, CURLMOPT_TIMERDATA, <void*>self)
        self.loop = loop
        self.timer_handle = None

    cdef void curl_perform_read(self, int fd):
        cdef int _running
        cdef CURLM* multi = self.multi
        with nogil:
            curl_multi_socket_action(multi, fd, CURL_CSELECT_IN, &_running)
        self.check_multi_info()

    cdef void curl_perform_write(self, int fd):
        cdef int _running
        cdef CURLM* multi = self.multi
        with nogil:
            curl_multi_socket_action(multi, fd, CURL_CSELECT_OUT, &_running)
        self.check_multi_info()

    cdef void timeout_expired(self):
        cdef int _running
        cdef CURLM* multi = self.multi
        with nogil:
            curl_multi_socket_action(multi, CURL_SOCKET_TIMEOUT, 0, &_running)
        self.check_multi_info()

    cdef void check_multi_info(self):
        cdef CURLMsg *message
        cdef int _pending
        cdef CURL *easy
        cdef _Response response
        cdef void *response_raw

        message = curl_multi_info_read(self.multi, &_pending)
        while message != NULL:
            if message.msg == CURLMSG_DONE:
                easy = message.easy_handle
                acurl_easy_getinfo_voidptr(easy, CURLINFO_PRIVATE, &response_raw)
                response = <_Response>response_raw
                if message.data.result == CURLE_OK:
                    response.future.set_result(response)
                else:
                    response.future.set_exception(AcurlError(f"curl failed with code {message.data.result} {curl_easy_strerror(message.data.result).decode('utf-8')}"))

                curl_multi_remove_handle(self.multi, easy)
            else:
                raise Exception("oops2")
            message = curl_multi_info_read(self.multi, &_pending)

    def session(self):
        return Session.__new__(Session, self)

    def __dealloc__(self):
        # FIXME: I (AWE) can't convince myself that this definitely doesn't
        # leak memory, because we might be tearing down the event loop before
        # we've called all the queued schedule_cleanup_curl_pointer events.
        # But this should be a rare case, so I'm not going to try to fix it
        # for now.
        curl_multi_cleanup(self.multi)



cdef extern from "<curl/curl.h>":
    # Forward decls
    ctypedef void CURL
    ctypedef void CURLM

    # Misc
    ctypedef int curl_socket_t
    ctypedef int (*curl_socket_callback)(CURL *easy, curl_socket_t s, int what, void *userp, void *socketp)
    ctypedef int (*curl_multi_timer_callback)(CURLM *multi, long timeout_ms, void *userp)
    ctypedef size_t (*curl_write_callback)(char *buffer, size_t size, size_t nitems, void *outstream)

    # Slist
    struct curl_slist:
        char *data
        curl_slist *next
    curl_slist *curl_slist_append(curl_slist *, const char *)
    void curl_slist_free_all(curl_slist *)

    # Multi
    ctypedef int CURLMcode
    ctypedef int CURLMoption

    cdef int CURLMOPT_SOCKETFUNCTION
    cdef int CURLMOPT_SOCKETDATA
    cdef int CURLMOPT_PIPELINING
    cdef int CURLMOPT_TIMERFUNCTION
    cdef int CURLMOPT_TIMERDATA
    cdef int CURLMOPT_MAXCONNECTS
    cdef int CURLMOPT_MAX_HOST_CONNECTIONS
    cdef int CURLMOPT_MAX_PIPELINE_LENGTH
    cdef int CURLMOPT_CONTENT_LENGTH_PENALTY_SIZE
    cdef int CURLMOPT_CHUNK_LENGTH_PENALTY_SIZE
    cdef int CURLMOPT_PIPELINING_SITE_BL
    cdef int CURLMOPT_PIPELINING_SERVER_BL
    cdef int CURLMOPT_MAX_TOTAL_CONNECTIONS
    cdef int CURLMOPT_PUSHFUNCTION
    cdef int CURLMOPT_PUSHDATA
    cdef int CURLMOPT_MAX_CONCURRENT_STREAMS
    cdef int CURLMOPT_LASTENTRY

    CURLM *curl_multi_init()
    CURLMcode curl_multi_socket_action(CURLM *multi_handle, curl_socket_t s, int ev_bitmask, int *running_handles) nogil
    CURLMsg *curl_multi_info_read(CURLM *multi_handle, int *msgs_in_queue)
    CURLMcode curl_multi_add_handle(CURLM *multi_handle, CURL *curl_handle)
    CURLMcode curl_multi_remove_handle(CURLM *multi_handle, CURL *curl_handle)
    CURLMcode curl_multi_cleanup(CURLM *multi_handle)

    # Curl
    ctypedef int CURLcode
    cdef union CURLMsgdata:
        void *whatever
        CURLcode result
    ctypedef int CURLMSG
    cdef struct CURLMsg:
        CURLMSG msg
        CURL *easy_handle
        CURLMsgdata data

    CURL *curl_easy_init()
    void curl_easy_cleanup(CURL *curl)

    # Misc enums
    cdef int CURL_POLL_NONE
    cdef int CURL_POLL_IN
    cdef int CURL_POLL_OUT
    cdef int CURL_POLL_INOUT
    cdef int CURL_POLL_REMOVE

    cdef int CURL_CSELECT_IN
    cdef int CURL_CSELECT_OUT
    cdef int CURL_CSELECT_ERR

    cdef int CURL_SOCKET_TIMEOUT

    cdef int CURLMSG_DONE

    # Share
    ctypedef void CURLSH
    ctypedef int CURLSHcode
    ctypedef int CURLSHoption
    cdef int CURLSHOPT_SHARE
    cdef int CURL_LOCK_DATA_COOKIE
    cdef int CURL_LOCK_DATA_DNS
    cdef int CURL_LOCK_DATA_SSL_SESSION

    CURLSH *curl_share_init()
    void curl_share_cleanup(CURLSH* share)

    # Options
    ctypedef int CURLoption
    cdef int CURLOPT_SHARE
    cdef int CURLOPT_URL
    cdef int CURLOPT_CUSTOMREQUEST
    cdef int CURLOPT_ENCODING
    cdef int CURLOPT_SSL_VERIFYPEER
    cdef int CURLOPT_SSL_VERIFYHOST
    cdef int CURLOPT_PRIVATE
    cdef int CURLOPT_HTTPHEADER
    cdef int CURLOPT_USERPWD
    cdef int CURLOPT_SSLKEY
    cdef int CURLOPT_SSLCERT
    cdef int CURLOPT_WRITEFUNCTION
    cdef int CURLOPT_WRITEDATA
    cdef int CURLOPT_HEADERFUNCTION
    cdef int CURLOPT_HEADERDATA
    cdef int CURLOPT_COOKIEFILE
    cdef int CURLOPT_COOKIELIST
    cdef int CURLOPT_POSTFIELDSIZE
    cdef int CURLOPT_POSTFIELDS

    # Info
    ctypedef int CURLINFO
    cdef int CURLINFO_PRIVATE
    cdef int CURLINFO_EFFECTIVE_URL
    cdef int CURLINFO_RESPONSE_CODE
    cdef int CURLINFO_TOTAL_TIME
    cdef int CURLINFO_NAMELOOKUP_TIME
    cdef int CURLINFO_CONNECT_TIME
    cdef int CURLINFO_COOKIELIST
    cdef int CURLINFO_APPCONNECT_TIME
    cdef int CURLINFO_PRETRANSFER_TIME
    cdef int CURLINFO_STARTTRANSFER_TIME
    cdef int CURLINFO_SIZE_UPLOAD
    cdef int CURLINFO_SIZE_DOWNLOAD
    cdef int CURLINFO_PRIMARY_IP
    cdef int CURLINFO_REDIRECT_URL

    # Error codes
    cdef int CURLE_OK

# Typed wrappers
cdef extern from "acurl_wrappers.h":
    # curl_multi_setopt
    CURLMcode acurl_multi_setopt_pointer(CURLM * multi_handle, CURLMoption option, void * param)
    CURLMcode acurl_multi_setopt_long(CURLM * multi_handle, CURLMoption option, long param)
    CURLMcode acurl_multi_setopt_socketcb(CURLM * multi_handle, CURLMoption option, curl_socket_callback param)
    CURLMcode acurl_multi_setopt_timercb(CURLM * multi_handle, CURLMoption option, curl_multi_timer_callback param)
    # curl_easy_getinfo
    CURLcode acurl_easy_getinfo_long(CURL *curl, CURLINFO info, long *data)
    CURLcode acurl_easy_getinfo_double(CURL *curl, CURLINFO info, double *data)
    CURLcode acurl_easy_getinfo_cstr(CURL *curl, CURLINFO info, char **data)
    CURLcode acurl_easy_getinfo_voidptr(CURL *curl, CURLINFO info, void **data)
    # curl_share_setopt
    CURLSHcode acurl_share_setopt_int(CURLSH *share, CURLSHoption option, int data)
    # curl_easy_setopt
    CURLcode acurl_easy_setopt_voidptr(CURL *easy, CURLoption option, void *data)
    CURLcode acurl_easy_setopt_cstr(CURL *easy, CURLoption option, const char *data)
    CURLcode acurl_easy_setopt_int(CURL *easy, CURLoption option, int data)
    CURLcode acurl_easy_setopt_writecb(CURL *easy, CURLoption option, curl_write_callback data)
    # curl_easy_strerror
    const char *curl_easy_strerror(CURLcode errornum)



#cython: language_level=3

import shlex
from urllib.parse import urlparse

cdef class Request:
    cdef readonly bytes method
    cdef readonly str url
    cdef tuple header_tuple
    cdef tuple cookie_tuple
    cdef readonly object auth
    cdef readonly object data
    cdef readonly object cert
    cdef tuple session_cookies
    cdef curl_slist* curl_headers

    def __cinit__(
        self,
        bytes method,
        str url,
        tuple header_tuple,
        tuple cookie_tuple,
        tuple auth,
        bytes data,
        tuple cert,
    ):
        self.method = method
        self.url = url
        self.header_tuple = header_tuple  # Tuple of byte strings
        self.cookie_tuple = cookie_tuple  # Tuple of byte strings
        self.auth = auth
        self.data = data
        self.cert = cert
        self.session_cookies = ()  # Tuple of byte strings
        self.curl_headers = NULL

    def __dealloc__(self):
        if self.curl_headers != NULL:
            curl_slist_free_all(self.curl_headers)

    cdef void store_session_cookies(self, CURLSH* shared):
        cdef CURL* curl = curl_easy_init()
        acurl_easy_setopt_voidptr(curl, CURLOPT_SHARE, shared)
        raw_cookies = tuple(parse_cookie_string(c) for c in acurl_extract_cookielist(curl))
        curl_easy_cleanup(curl)
        session_cookies = tuple(
            c.format()
            for c in raw_cookies
            if urlparse(self.url).hostname.lower() == c.domain.lower()
        )
        self.session_cookies = session_cookies

    @property
    def headers(self):
        d = dict(header.split(b": ", 1) for header in self.header_tuple)
        return {
            k.decode("utf-8"): v.decode("utf-8")
            for k, v in d.items()
        }

    @property
    def cookies(self):
        request_cookies = ()
        for c in self.cookie_tuple:
            request_cookies += (parse_cookie_string(c.decode("utf-8")),)
        session_cookies = tuple(parse_cookie_string(c.decode("utf-8")) for c in self.session_cookies)

        return cookie_seq_to_cookie_dict(request_cookies + session_cookies)

    def to_curl(self):
        cdef object data
        cdef list args = ["curl", "-X", self.method.decode('ascii')]
        if self.data is not None:
            data = self.data
            if hasattr(data, "decode"):
                data = data.decode("utf-8")
            args.append("-d " + shlex.quote(data))
        if len(self.headers) > 0:
            args.append(
                " ".join(
                    ("-H " + shlex.quote(k + ": " + v) for k, v in self.headers.items())
                )
            )
        if len(self.cookies) > 0:
            args.append(
                "--cookie " + shlex.quote(
                    ";".join((f"{k}={v}" for k, v in self.cookies.items()))
                )
            )
        if self.auth is not None:
            args.append("--user " + shlex.quote(f"{self.auth[0]}:{self.auth[1]}"))
        args.append(shlex.quote(self.url))
        return " ".join(args)



import time

import pytest

import acurl_ng
from acurl_ng import Cookie


@pytest.mark.skip
def test_cookie_zero_expiry():
    # FIXME: what's the right behavior here?
    c = acurl_ng._Cookie(
        False, "foo.com", False, "/bar", False, 0, "my_cookie", "my_value"
    )
    assert not c.has_expired


def test_cookie_not_expired():
    c = acurl_ng._Cookie(
        False,
        "foo.com",
        False,
        "/bar",
        False,
        time.time() + 200,
        "my_cookie",
        "my_value",
    )
    assert not c.has_expired


def test_cookie_has_expired():
    c = acurl_ng._Cookie(
        False, "foo.com", False, "/bar", False, 1, "my_cookie", "my_value"
    )
    assert c.has_expired


def test_cookie_format():
    c = acurl_ng._Cookie(
        False, "foo.com", False, "/bar", False, 0, "my_cookie", "my_value"
    )
    assert c.format() == b"foo.com\tFALSE\t/bar\tFALSE\t0\tmy_cookie\tmy_value"


def test_parse_cookie_string():
    c = acurl_ng.parse_cookie_string(
        "foo.com\tFALSE\t/bar\tFALSE\t0\tmy_cookie\tmy_value"
    )
    assert not c.http_only
    assert c.domain == "foo.com"
    assert not c.include_subdomains
    assert c.path == "/bar"
    assert not c.is_secure
    assert c.expiration == 0
    assert c.name == "my_cookie"
    assert c.value == "my_value"


def test_parse_cookie_string_with_true():
    c = acurl_ng.parse_cookie_string("foo.com\tTRUE\t/bar\tTRUE\t0\tmy_cookie\tmy_value")
    assert not c.http_only
    assert c.domain == "foo.com"
    assert c.include_subdomains
    assert c.path == "/bar"
    assert c.is_secure
    assert c.expiration == 0
    assert c.name == "my_cookie"
    assert c.value == "my_value"


def test_parse_cookie_string_http_only():
    c = acurl_ng.parse_cookie_string(
        "#HttpOnly_foo.com\tFALSE\t/bar\tFALSE\t0\tmy_cookie\tmy_value"
    )
    assert c.http_only
    assert c.domain == "foo.com"
    assert not c.include_subdomains
    assert c.path == "/bar"
    assert not c.is_secure
    assert c.expiration == 0
    assert c.name == "my_cookie"
    assert c.value == "my_value"


def test_session_cookie_for_url_with_cookie_instance():
    c = Cookie("sky.com", "foo", "bar")
    cookie = acurl_ng.session_cookie_for_url(
        url="https://anotherdomain.com", name="baz", value=c
    )
    assert cookie.domain == "sky.com"
    assert cookie.name == "foo"
    assert cookie.value == "bar"


def test_session_cookie_for_url_with_string_value():
    cookie = acurl_ng.session_cookie_for_url(
        url="https://sky.com", name="foo", value="bar"
    )
    assert cookie.domain == ".sky.com"
    assert cookie.name == "foo"
    assert cookie.value == "bar"



#cython: language_level=3

from libc.time cimport time
from urllib.parse import urlparse


cdef class Cookie:
    cdef readonly str domain
    cdef readonly str name
    cdef readonly str value

    def __cinit__(self, domain, name, value):
        self.domain = domain
        self.name = name
        self.value = value

cdef class _Cookie:
    cdef readonly bint http_only
    cdef readonly str domain
    cdef readonly bint include_subdomains
    cdef readonly str path
    cdef readonly bint is_secure
    cdef readonly int expiration
    cdef readonly str name
    cdef readonly str value

    def __cinit__(
        self,
        bint http_only,
        str domain,
        bint include_subdomains,
        str path,
        bint is_secure,
        int expiration,
        str name,
        str value
    ):
        self.http_only = http_only
        self.domain = domain
        self.include_subdomains = include_subdomains
        self.path = path
        self.is_secure = is_secure
        self.expiration = expiration
        self.name = name
        self.value = value

    @property
    def has_expired(self):
        return self.expiration != 0 and time(NULL) > self.expiration

    cpdef bytes format(self):
        cdef array.array bits = array.array('B', [])
        if self.http_only:
            bits.frombytes(b"#HttpOnly_")
        bits.frombytes(self.domain.encode())
        bits.append(9)  # Tab
        bits.frombytes(b"TRUE" if self.include_subdomains else b"FALSE")
        bits.append(9)
        bits.frombytes(self.path.encode())
        bits.append(9)
        bits.frombytes(b"TRUE" if self.is_secure else b"FALSE")
        bits.append(9)
        bits.frombytes(str(self.expiration).encode())
        bits.append(9)
        bits.frombytes(self.name.encode())
        bits.append(9)
        bits.frombytes(self.value.encode())
        return bits.tobytes()

cpdef _Cookie session_cookie_for_url(
    str url,
    str name,
    object value,
    bint http_only=False,
    bint include_subdomains=True,
    bint is_secure=False,
    bint include_url_path=False,
):
    cdef str scheme, netloc, path, params, query, fragment
    scheme, netloc, path, params, query, fragment = urlparse(url)
    if not include_url_path:
        path = "/"

    cdef bint is_type_cookie
    if isinstance(value, Cookie):
        is_type_cookie = True
    elif isinstance(value, str):
        is_type_cookie = False
    else:
        raise ValueError("cookie value must be string or Cookie class")

    # TODO do we need to sanitize netloc for IP and ports?
    return _Cookie(
        http_only,
        value.domain if is_type_cookie else "." + netloc.split(":")[0],
        include_subdomains,
        path,
        is_secure,
        0,
        value.name if is_type_cookie else name,
        value.value if is_type_cookie else value,
    )

cpdef _Cookie parse_cookie_string(str cookie_string):
    cdef bint http_only
    cdef str domain, include_subdomains, path, is_secure, expiration, name, value

    cookie_string = cookie_string.strip()
    if cookie_string.startswith("#HttpOnly_"):  # FIXME: optimize?
        http_only = True
        cookie_string = cookie_string[10:]
    else:
        http_only = False
    cdef list parts = cookie_string.split("\t")
    if len(parts) == 6:
        domain, include_subdomains, path, is_secure, expiration, name = parts
        value = ""
    else:
        domain, include_subdomains, path, is_secure, expiration, name, value = parts
    return _Cookie(
        http_only,
        domain,
        include_subdomains == "TRUE",
        path,
        is_secure == "TRUE",
        int(expiration),
        name,
        value,
    )

cdef dict cookie_seq_to_cookie_dict(tuple cookie_list):
    cdef int i
    cdef dict d = {}
    cdef _Cookie cookie
    for i in range(len(cookie_list)):
        cookie = cookie_list[i]
        d[cookie.name] = cookie.value
    return d



from unittest.mock import Mock

import pytest
from helpers_ng import create_request

import acurl_ng


def test_request_headers():
    r = create_request("GET", "http://foo.com", headers=("Foo: bar", "Baz: quux"))
    assert "Foo" in r.headers
    assert r.headers["Foo"] == "bar"
    assert "Baz" in r.headers
    assert r.headers["Baz"] == "quux"


def test_request_cookies():
    session_mock = Mock()
    session_mock.get_cookie_list.return_value = ()
    r = create_request(
        "GET",
        "http://foo.com",
        cookies=(
            acurl_ng._Cookie(
                False, "foo.com", True, "/", False, 0, "foo", "bar"
            ).format(),
            acurl_ng._Cookie(
                False, "foo.com", True, "/", False, 0, "my_cookie", "my_value"
            ).format(),
        ),
    )
    assert "my_cookie" in r.cookies
    assert r.cookies["my_cookie"] == "my_value"
    assert "foo" in r.cookies
    assert r.cookies["foo"] == "bar"


@pytest.mark.asyncio
async def test_request_e2e(httpbin, acurl_session_ng):
    r = await acurl_session_ng.get(
        httpbin.url + "/get", headers={"Foo": "bar"}, cookies={"baz": "quux"}
    )
    assert r.request.cookies == {"baz": "quux"}
    assert r.request.headers == {"Foo": "bar"}


@pytest.mark.asyncio
async def test_request_cookies_from_previous(httpbin, acurl_session_ng):
    await acurl_session_ng.get(httpbin.url + "/cookies/set?name=value")
    r = await acurl_session_ng.get(httpbin.url + "/get", cookies={"foo": "bar"})
    assert r.request.cookies == {"name": "value", "foo": "bar"}


@pytest.mark.asyncio
@pytest.mark.slow
async def test_request_cookies_from_previous_excludes_other_domains(
    httpbin, acurl_session_ng
):
    await acurl_session_ng.get(httpbin.url + "/cookies/set?name=value")
    # FIXME: we want to set a cookie for another domain.  There's no easy way
    # to get another domain set up loally, so we (as an exception) go out to
    # the network for this test.
    await acurl_session_ng.get("https://httpbin.org/cookies/set?foo=bar")
    r = await acurl_session_ng.get(httpbin.url + "/get")
    assert r.request.cookies == {"name": "value"}



import asyncio
from urllib.parse import urlencode

import pytest

import acurl_ng


async def session():
    el = acurl_ng.CurlWrapper(asyncio.get_running_loop())
    return el.session()


@pytest.mark.asyncio
async def test_get(httpbin):
    s = await session()
    r = await s.get(httpbin.url + "/ip")
    assert r.status_code == 200
    assert isinstance(r.headers, dict)
    # FIXME: is this a method in the requests api?
    assert isinstance(r.json(), dict)


@pytest.mark.asyncio
async def test_cookies(httpbin, acurl_session_ng):
    r = await acurl_session_ng.get(httpbin.url + "/cookies/set?name=value")
    assert r.cookies == {"name": "value"}


@pytest.mark.asyncio
async def test_session_cookies(httpbin, acurl_session_ng):
    await acurl_session_ng.get(httpbin.url + "/cookies/set?name=value")
    cookies = acurl_session_ng.cookies()
    assert cookies == {"name": "value"}
    acurl_session_ng.erase_all_cookies()
    cookie_list = acurl_session_ng.cookies()
    assert len(cookie_list) == 0


@pytest.mark.asyncio
async def test_session_cookies_sent_on_subsequent_request(httpbin):
    s = await session()
    await s.get(httpbin.url + "/cookies/set?name=value")
    resp = await s.get(httpbin.url + "/cookies")
    data = resp.json()
    assert len(data) == 1
    assert data["cookies"] == {"name": "value"}


@pytest.mark.asyncio
async def test_set_cookies(httpbin):
    s = await session()
    await s.get(httpbin.url + "/cookies/set?name=value")
    r = await s.get(httpbin.url + "/cookies/set?name2=value", cookies={"name3": "value"})
    assert r.cookies == {"name": "value", "name2": "value", "name3": "value"}


@pytest.mark.asyncio
async def test_basic_auth(httpbin):
    s = await session()
    r = await s.get(httpbin.url + "/basic-auth/user/password", auth=("user", "password"))
    assert r.status_code == 200


@pytest.mark.asyncio
async def test_failed_basic_auth(httpbin):
    s = await session()
    r = await s.get(
        httpbin.url + "/basic-auth/user/password", auth=("notuser", "notpassword")
    )
    assert r.status_code == 401


@pytest.mark.asyncio
async def test_redirect(httpbin):
    s = await session()
    url = httpbin.url + "/ip"
    r = await s.get(httpbin.url + "/redirect-to?" + urlencode({"url": url}))
    assert r.url == url



from acurl_ng import CaseInsensitiveDict


def test_case_insensitive_constructor():
    # Expect it to lower keys on return
    my_dict = CaseInsensitiveDict({"Foo": "Bar", "baz": "Qux"})
    assert list(my_dict.keys()) == ["foo", "baz"]


def test_case_insensitive_get():
    my_dict = CaseInsensitiveDict({"Foo": "Bar"})
    assert my_dict.get("Foo") == "Bar"
    assert my_dict.get("foo") == "Bar"


def test_case_insensitive_get_item():
    my_dict = CaseInsensitiveDict({"Foo": "Bar"})
    assert my_dict["Foo"] == "Bar"
    assert my_dict["foo"] == "Bar"


def test_case_insensitive_set_item():
    my_dict = CaseInsensitiveDict()
    my_dict["Foo"] = "Bar"
    assert my_dict["Foo"] == "Bar"
    assert my_dict["foo"] == "Bar"
    my_dict["foo"] = "Bar"
    assert my_dict["Foo"] == "Bar"
    assert my_dict["foo"] == "Bar"


def test_case_insensitive_del_item():
    my_dict = CaseInsensitiveDict({"Foo": "Bar", "baz": "Qux"})
    del my_dict["Foo"]
    assert list(my_dict.keys()) == ["baz"]
    del my_dict["Baz"]
    assert list(my_dict.keys()) == []


def test_case_insensitive_contains():
    my_dict = CaseInsensitiveDict({"Foo": "Bar"})
    assert "Foo" in my_dict
    assert "foo" in my_dict


def test_case_insensitive_pop():
    my_dict = CaseInsensitiveDict({"Foo": "Bar", "baz": "Qux"})
    assert my_dict.pop("Foo") == "Bar"
    assert my_dict.pop("Baz") == "Qux"
    assert list(my_dict.keys()) == []


def test_case_insensitive_popitem():
    my_dict = CaseInsensitiveDict({"Foo": "Bar", "Baz": "Qux"})
    assert my_dict.popitem() == ("baz", "Qux")
    assert list(my_dict.keys()) == ["foo"]


def test_case_insensitive_setdefault():
    my_dict = CaseInsensitiveDict({"Foo": "Bar"})
    assert my_dict.setdefault("Foo") == "Bar"
    assert my_dict.setdefault("foo") == "Bar"
    assert my_dict.setdefault("Baz", "Qux") == "Qux"
    assert my_dict.setdefault("baz") == "Qux"



import os

from Cython.Build import cythonize
from Cython.Compiler.Options import get_directive_defaults
from setuptools import Extension, setup

# https://stackoverflow.com/a/28301932/463500
if "IS_TOX_BUILD" in os.environ:
    directive_defaults = get_directive_defaults()
    directive_defaults["linetrace"] = True
    # directive_defaults['binding'] = True
    kwargs = {"define_macros": [("CYTHON_TRACE", "1")]}
else:
    kwargs = {}

extensions = [Extension("acurl_ng", ["src/acurl.pyx"], libraries=["curl"], **kwargs)]


def local_scheme(version):
    return ""


setup(
    ext_modules=cythonize(
        extensions,
        gdb_debug=True,
        compiler_directives={
            "warn.undeclared": True,
            "warn.unreachable": True,
            "warn.maybe_uninitialized": True,
            "warn.unused": True,
            "warn.unused_arg": True,
            "warn.unused_result": True,
        },
    ),
    setup_requires=["cython"],
    # use_scm_version={"root": "..", "relative_to": __file__, "local_scheme": local_scheme},
)



import pytest
from werkzeug.datastructures import Headers
from werkzeug.wrappers import Response


@pytest.mark.asyncio
async def test_response_headers(httpserver, acurl_session_ng):
    hdrs = Headers()
    hdrs.add("Foo", "bar")
    hdrs.add("Baz", "quux")
    hdrs.add("baz", "quuz")
    httpserver.expect_request("/foo").respond_with_response(
        Response(response="", status=200, headers=hdrs)
    )
    r = await acurl_session_ng.get(httpserver.url_for("/foo"))
    assert "Foo" in r.headers
    assert r.headers["Foo"] == "bar"
    assert r.headers["foo"] == "bar"
    assert "Baz" in r.headers
    assert r.headers["Baz"] == "quux, quuz"
    assert r.headers["baz"] == "quux, quuz"


async def connected_cb(body, reader, writer):
    writer.write(body)
    await writer.drain()
    writer.close()
    await writer.wait_closed()
    await reader.read(-1)


@pytest.mark.asyncio
async def test_response_headers_with_HTTP_100(httpserver, acurl_session_ng):
    hdrs = Headers()
    hdrs.add("Foo", "bar")
    httpserver.expect_request("/foo").respond_with_response(
        Response(response="", status=200, headers=hdrs)
    )
    r = await acurl_session_ng.get(
        httpserver.url_for("/foo"), headers={"Expect": "100-continue"}
    )

    assert "Foo" in r.headers
    assert r.headers["Foo"] == "bar"


@pytest.mark.asyncio
async def test_response_cookies(httpserver, acurl_session_ng):
    hdrs = Headers()
    hdrs.add("Set-Cookie", "foo=bar")
    hdrs.add("Set-Cookie", "quux=xyzzy")
    httpserver.expect_request("/foo").respond_with_response(
        Response(response="", status=200, headers=hdrs)
    )
    r = await acurl_session_ng.get(httpserver.url_for("/foo"))
    assert r.cookies == {"foo": "bar", "quux": "xyzzy"}



from collections import defaultdict


class _CaseInsensitiveDict:
    # Based on https://stackoverflow.com/a/32888599 but tweaked for python3
    @staticmethod
    def _k(key):
        return key.lower() if isinstance(key, str) else key

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._convert_keys()

    def __getitem__(self, key):
        return super().__getitem__(self._k(key))

    def __setitem__(self, key, value):
        super().__setitem__(self._k(key), value)

    def __delitem__(self, key):
        return super().__delitem__(self._k(key))

    def __contains__(self, key):
        return super().__contains__(self._k(key))

    def pop(self, key, *args, **kwargs):
        return super().pop(self._k(key), *args, **kwargs)

    def popitem(self, *args, **kwargs):
        return super().popitem(*args, **kwargs)

    def get(self, key, *args, **kwargs):
        return super().get(self._k(key), *args, **kwargs)

    def setdefault(self, key, *args, **kwargs):
        return super().setdefault(self._k(key), *args, **kwargs)

    def update(self, E=None, **F):
        super().update(self.__class__(E, **F))

    def _convert_keys(self):
        for k in list(self.keys()):
            v = super().pop(k)
            self.__setitem__(k, v)


class CaseInsensitiveDict(_CaseInsensitiveDict, dict):
    pass


class CaseInsensitiveDefaultDict(_CaseInsensitiveDict, defaultdict):
    pass
